from pydantic import BaseModel,Field
from typing import ForwardRef,Union
from typing import List
from zt_backend.runner.user_state import UserContext

Row = ForwardRef('Row')
Column = ForwardRef('Column')

class Column(BaseModel):
    components: List[Union[str,"Row"]] = Field([], description="List of component IDs that belong to this column")
    width: Union[int,bool] = Field(False)
    

class Row(BaseModel):
    components: List[Union[str,Column]] = Field([], description="List of columns that belong to this row")


class Layout(BaseModel):
    rows: List[Row] = Field([], description="List of rows in this layout")
    columns: List[Column] = Field([], description="List of columns in this layout")

    def __init__(self, **data):
        super().__init__(**data)
        execution_state = UserContext.get_state()
        if execution_state and execution_state.context_globals['exec_mode']:
            execution_state.current_cell_layout.append(self)


from pydantic import Field, validator,field_validator
from zt_backend.models.components.zt_component import ZTComponent
from zt_backend.models.validations import validate_color
from typing import List, Union, Optional
from zt_backend.runner.user_state import UserContext

class SelectBox(ZTComponent):
    """A class for SelectBox components inheriting from ZTComponent."""
    
    component: str = Field("v-select", description="Vue component name.")
    items: List[Union[str, int]] = Field(..., description="Options for the select box. Can be a list of strings or integers.")
    value: Union[str, int, None] = Field(None, description="Selected option for the select box. Can be a string or integer.")
    label: Optional[str] = Field(None, description="Label for the select box.")
    multiple: Optional[bool] = Field(None, description="Determines if multiple selections are allowed.")
    dense: Optional[bool] = Field(None, description="Determines if the select box is dense.")
    outlined: Optional[bool] = Field(None, description="Determines if the select box has an outlined style.")
    clearable: Optional[bool] = Field(None, description="Determines if the select box has a clearable option.")
    disabled: Optional[bool] = Field(None, description="Determines if the select box is disabled.")
    readonly: Optional[bool] = Field(None, description="Determines if the select box is read-only.")
    color: Optional[str] = Field(None, pre=True, description="Color of the range slider. Can be custom or standard Material color.")
    triggerEvent: str = Field('update:modelValue',description="Trigger event for when to run the slider")

    @field_validator('color')
    def validate_color(cls, color):
        return validate_color(color)

    @validator('value', always=True) #TODO: debug and replace with field validator
    def get_value_from_global_state(cls, value, values):
        id = values['id'] # Get the id if it exists in the field values
        execution_state = UserContext.get_state()
        try:
            if execution_state and id and id in execution_state.component_values:  # Check if id exists in global_state
                return execution_state.component_values[id]  # Return the value associated with id in global_state
        except Exception as e:
            e
        return value  # If id doesn't exist in global_state, return the original value


from pydantic import Field
from zt_backend.models.components.zt_component import ZTComponent
from typing import Dict
from plotly.graph_objs import Figure
import pandas as pd
import numpy as np

def convert_special_types(obj):
    """Recursively convert special types like DataFrame and NumPy array to list."""
    if isinstance(obj, dict):
        for key, value in obj.items():
            if isinstance(value, (pd.DataFrame, np.ndarray)):
                obj[key] = value.tolist()
            else:
                convert_special_types(value)
    elif isinstance(obj, list):
        for i, item in enumerate(obj):
            if isinstance(item, (pd.DataFrame, np.ndarray)):
                obj[i] = item.tolist()
            else:
                convert_special_types(item)

class PlotlyComponent(ZTComponent):
    component: str = Field("plotly-plot", description="Vue component name for Plotly.")
    figure: Dict = Field(..., description="Plotly figure object.")
    layout: dict = Field(..., description="Layout for Plotly plot.")
    id: str = Field(..., description="ID for the component")
    
    @classmethod
    def from_figure(cls, figure: Figure, id: str):
        """Creates a PlotlyComponent instance from a Plotly figure object."""
        figure.update_layout(template = 'plotly_dark')
        # Ensure the input is a Plotly figure
        if not isinstance(figure, Figure):
            raise ValueError("Input must be a Plotly Figure")
        # Convert the figure to JSON-serializable dictionary
        figure_json = figure.to_plotly_json()
        figure_data = figure_json.get('data', [])
        layout_data = figure_json['layout']
        #print(layout_data)
        # Convert any special types like DataFrame or NumPy array to list
        for trace in figure_data:
            convert_special_types(trace)
        return cls(figure=figure_json, layout=layout_data, id=id)
    
    def to_json(self):
        """Converts the Plotly figure to a JSON-serializable dictionary."""
        return {
            "component": self.component,
            "figure": self.figure,
            "layout": self.layout,
            "id": self.id
        }


from pydantic import Field, field_validator
from zt_backend.models.components.zt_component import ZTComponent
import base64

class Image(ZTComponent):
    """A class for Image components inheriting from ZTComponent."""
    component: str = Field("v-img", description="Vue component name.")
    src: str = Field(..., description="Source URL or Path of the image.")
    alt: str = Field("", description="Alternative text for the image.")
    width: int = Field(100, description="Width of the image.")
    height: int = Field(100, description="Height of the image.")

    @field_validator('src')
    def validate_src(cls, src):
        if src.endswith(('.png', ".jpg", ".jpeg")) and not src.startswith('http'):
            try:
                buffer=open(src, 'rb')
                buffer.seek(0)
                b64_img = base64.b64encode(buffer.read()).decode('utf-8')
                if src.endswith((".jpg", ".jpeg")):
                    return f"data:image/jpeg;base64,{b64_img}"
                elif src.endswith(".png"):
                    return f"data:image/png;base64,{b64_img}"
                else:
                    print("Image file type is not supported")
            except Exception as e:
                print("Image not found: ", e)
        return src


from pydantic import BaseModel, Field, field_validator
from zt_backend.runner.user_state import UserContext

class ZTComponent(BaseModel):
    id: str = Field(description="Unique id for a component")
    variable_name: str = Field('', description="Optional variable name associated with a component")

    def __init__(self, **data):
        super().__init__(**data)
        execution_state = UserContext.get_state()
        if execution_state and execution_state.context_globals['exec_mode']:
            execution_state.current_cell_components.append(self)

    @field_validator('id', mode='before')
    def validate_unique_component_id(cls, id):
        execution_state = UserContext.get_state()
        if execution_state and execution_state.context_globals['exec_mode'] and id in execution_state.created_components:
            raise Exception("Component ID is not unique")
        elif execution_state and execution_state.context_globals['exec_mode']:
            execution_state.created_components.append(id)
            return id
        else:
            return id


from typing import OrderedDict, List, Dict, Tuple, Any
import astroid
from zt_backend.models.request import Request,Cell,CodeDict
import duckdb
import uuid
import re
import logging
import traceback
from zt_backend.config import settings


logger = logging.getLogger("__name__")

def get_imports(module) -> List[str]:
    import_froms = [node.names[0][1] or node.names[0][0] for node in module.nodes_of_class(astroid.ImportFrom)]
    imports = [node.names[0][1] or node.names[0][0] for node in module.nodes_of_class(astroid.Import)]
    return import_froms + imports

def get_functions(module) -> Tuple[List[str], List[str]]:
    function_names = []
    argument_names = []
    for function_def in module.nodes_of_class(astroid.FunctionDef):
        function_names.append(function_def.name)
        argument_names.extend([arg.name for arg in function_def.args.args])
    return function_names, argument_names

def get_defined_names(module) -> List[str]:
    defined_names = []
    for defnode in module.nodes_of_class(astroid.Assign):
        for target in defnode.targets:
            if hasattr(target, 'name'):  # Directly has a name (e.g., AssignName)
                defined_names.append(target.name)
            elif isinstance(target, astroid.Subscript):  # Is a subscript
                if hasattr(target.value, 'name'):
                    defined_names.append(target.value.name)    
    func_def_names = [arg.name for func in module.nodes_of_class(astroid.FunctionDef) for arg in func.args.args]
    return list(set(defined_names) - set(func_def_names))

def get_loaded_modules(module,all_imports) -> List[str]:
    try:
        return list(set([node.expr.name for node in module.nodes_of_class(astroid.Attribute) if hasattr(node.expr, 'name') and node.expr.name in all_imports]))
    except Exception as e:
        logger.error("Error getting loaded modules: %s", traceback.format_exc())
        return []

def get_loaded_names(module, defined_names) -> List[str]:
    function_names, function_arguments = get_functions(module)
    return [usenode.name for usenode in module.nodes_of_class(astroid.Name) if usenode.name not in function_arguments]

def generate_sql_code(cell, uuid_value, db_file='zt_db.db'):
    """Generate SQL code for the given cell."""
    
    # Common import statements
    base_code = "import duckdb\nimport zero_true as zt"

    # Initialize the DuckDB database connection to persist tables
    db_init = f"conn = duckdb.connect('{db_file}')"

    # Extract all placeholders from the SQL string
    placeholders = re.findall(r"\{(.*?)\}", cell.code)
    
    # Replace placeholders with "?" for SQL parameterization
    parametrized_sql = re.sub(r"\{(.*?)\}", "?", cell.code)
    
    # SQL code execution
    sql_execution = f'conn.execute("""{parametrized_sql}""", [{", ".join(placeholders)}]).df()'

    if cell.variable_name:
        # If variable_name is provided, use it for assignment
        variable_assignment = f"{cell.variable_name} = {sql_execution}"

        # Convert the result to a custom DataFrame
        data_frame_conversion = f"zt.DataFrame.from_dataframe(id='{uuid_value}', df={cell.variable_name})"
        
        full_code = f"{base_code}\n{db_init}\n{variable_assignment}\n{data_frame_conversion}"
        
    else:
        # If variable_name is not provided, directly use the SQL execution
        if settings.run_mode == 'app':
            data_frame_conversion = ''
        else:
            data_frame_conversion = f"zt.DataFrame.from_dataframe(id='{uuid_value}', df={sql_execution})"
        
        full_code = f"{base_code}\n{db_init}\n{data_frame_conversion}"

    return full_code

def parse_cells(request: Request) -> CodeDict:
    cell_dict = {}
    all_imports = []
    for cell in [c for c in request.cells if c.cellType in ['code', 'sql']]:
        table_names=[]
        if cell.cellType=='sql' and cell.code:
            try:
                table_names = duckdb.get_table_names(re.sub(r'\{.*?\}', '1', cell.code))
            except Exception as e:
                print(e)
            uuid_value = str(uuid.uuid4())
            cell.code = generate_sql_code(cell, uuid_value)
        
        try:
            module = astroid.parse(cell.code)
            all_imports += get_imports(module)
            function_names, function_arguments = get_functions(module)
            defined_names = get_defined_names(module) + function_names
            loaded_names = [name for name in get_loaded_names(module, defined_names) + get_loaded_modules(module,all_imports) + list(table_names) if name not in get_imports(module)]
            cell_dict[cell.id] = Cell(**{
                'code': cell.code,
                'defined_names': defined_names,
                'imported_modules':get_imports(module),
                'loaded_modules':get_loaded_modules(module,all_imports),
                'loaded_names': list(set(loaded_names))})
        except Exception as e:
            logger.error("Error while parsing cells, returning empty names lists: %s", traceback.format_exc())
            cell_dict[cell.id] = Cell(**{
                'code': cell.code,
                'defined_names': [],
                'loaded_names': []})
    
    return CodeDict(cells=cell_dict)


def build_dependency_graph(cell_dict: Dict[int, Dict[str, Any]]) -> Dict[int, Dict[str, Any]]:
    """
    Builds a dependency graph from a cell dictionary. Each node in the graph represents a cell,
    and edges represent dependencies between cells based on loaded and defined names.
    """
    # Initialize previous components dictionary
    prev_components = {}

    # Add child and parent relationships
    graph = add_child_cells(cell_dict, prev_components)

    return graph

def find_child_cells(cell: Cell, code_dictionary: CodeDict, idx: int) -> List[str]:
    child_cells = []
    names = cell.defined_names
    for next_key in list(code_dictionary.cells.keys())[idx + 1:]:
        next_cell = code_dictionary.cells[next_key]
        next_loaded_names = next_cell.loaded_names
        next_loaded_modules = next_cell.loaded_modules
        if set(names).intersection(set(next_loaded_names)-set(next_loaded_modules)):
            child_cells.append(next_key)
    return child_cells

def add_parent_cells(code_dictionary: CodeDict) -> CodeDict:
    for key in list(code_dictionary.cells.keys()):
        cell = code_dictionary.cells[key]
        cell.parent_cells = find_parent_cells(cell,key, code_dictionary)
    return code_dictionary

def find_parent_cells(cell: Cell,key, code_dictionary: CodeDict) -> List[str]:
    parent_cells = []
    names = cell.loaded_names+cell.loaded_modules
    for prev_key in list(code_dictionary.cells.keys()):
        if prev_key == key:
            break
        prev_cell = code_dictionary.cells[prev_key]
        prev_defined_names = prev_cell.defined_names+prev_cell.loaded_names+prev_cell.loaded_modules+prev_cell.imported_modules
        if set(names).intersection(set(prev_defined_names)):
            parent_cells.append(prev_key)
    return parent_cells



def add_child_cells(code_dictionary: CodeDict, prev_components: Dict[str, Any]) -> Dict[str, Any]:
    for idx, key in enumerate(list(code_dictionary.cells.keys())):
        cell = code_dictionary.cells[key]
        cell.child_cells = find_child_cells(cell, code_dictionary, idx)
        cell.previous_child_cells = prev_components.get(key, {}).get('child_cells', [])
    return add_parent_cells(code_dictionary)

def print_astroid_tree(code):
    module = astroid.parse(code)
    print(module.repr_tree())

def find_downstream_cells(code_dictionary: CodeDict, start_cell_id, visited=None):
    if visited is None:
        visited = set()
        
    if start_cell_id in visited:
        return []
    
    visited.add(start_cell_id)
    
    downstream_cells = []
    
    for cell_id, cell_data in code_dictionary.cells.items():
        if start_cell_id in cell_data.parent_cells:
            downstream_cells.append(cell_id)
            downstream_cells.extend(find_downstream_cells(code_dictionary, cell_id, visited))
    return list(OrderedDict.fromkeys(downstream_cells))




from zt_backend.models.components.zt_component import ZTComponent
from zt_backend.models.components.slider import Slider
from zt_backend.models.components.text_input import TextInput
from zt_backend.models.components.text_area_input import TextArea
from zt_backend.models.components.range_slider import RangeSlider
from zt_backend.models.components.selectbox import SelectBox
from zt_backend.models.components.multiselect import MultiSelectBox
from zt_backend.models.components.button import Button
from zt_backend.models.components.number_input import NumberInput
from zt_backend.models.components.image import Image
from zt_backend.models.components.text import Text
from zt_backend.models.components.dataframe import DataFrame
from zt_backend.models.components.layout import Column,Row,Layout
from zt_backend.models.components.matplotlib import Matplotlib
from zt_backend.models.components.plotly import  PlotlyComponent
from zt_backend.models.components.autocomplete import  Autocomplete
from zt_backend.models.components.card import Card


from typing import OrderedDict
from matplotlib.pylab import f
from zt_backend.runner.user_state import UserState
from zt_backend.models import request, notebook, response
from dictdiffer import diff
import logging
import duckdb
import uuid
import os
import jedi
import traceback
import site
import json
import rtoml
from threading import Timer

logger = logging.getLogger("__name__")
notebook_db_dir =  site.USER_SITE+'/.zero_true/'
notebook_db_path = notebook_db_dir+'notebook.db'
codeCell = notebook.CodeCell(
                id=str(uuid.uuid4()),
                code='',
                components=[],
                variable_name='',
                output='',
                cellType='code'
            )
zt_notebook = notebook.Notebook(userId='', cells=OrderedDict([(codeCell.id, codeCell)]))

def debounce(wait):
    """ Decorator that will postpone a functions
        execution until after wait seconds
        have elapsed since the last time it was invoked. """
    def decorator(fn):
        def debounced(*args, **kwargs):
            def call_it():
                fn(*args, **kwargs)
            try:
                debounced.t.cancel()
            except(AttributeError):
                pass
            debounced.t = Timer(wait, call_it)
            debounced.t.start()
        return debounced
    return decorator


def get_notebook(id=''):
    global zt_notebook
    if id!='':
        try:
            logger.debug("Getting notebook from db with id %s", id)
            #get notebook from the database
            zt_notebook = get_notebook_db(id)
            return
        except Exception as e:
            logger.debug("Error when getting notebook %s from db: %s", id, traceback.format_exc())

    try:
        logger.debug("Notebook id is empty")            
        # If it doesn't exist in the database, load it from the TOML file
        with open('notebook.ztnb', "r") as project_file:
            toml_data = rtoml.loads(project_file.read().replace('\\','\\\\'))

        try:
            #get notebook from the database
            zt_notebook = get_notebook_db(toml_data['notebookId'])
            logger.debug("Notebook retrieved from db with id %s", toml_data['notebookId'])
            return
        except Exception as e:
            logger.debug("Error loading notebook with id %s from db: %s", toml_data['notebookId'], traceback.format_exc())
            pass
        # Convert TOML data to a Notebook object
        notebook_data = {
            'notebookId' : toml_data['notebookId'],
            'userId' : '',
            'cells': {
                cell_id: notebook.CodeCell(id=cell_id, **cell_data, output="")
                for cell_id, cell_data in toml_data['cells'].items()
            }
        }
        zt_notebook = notebook.Notebook(**notebook_data)
        new_notebook = zt_notebook.model_dump_json()
        conn = duckdb.connect(notebook_db_path)
        create_table_query = f"CREATE TABLE IF NOT EXISTS '{zt_notebook.notebookId}' (id STRING PRIMARY KEY, notebook STRING)"
        conn.execute(create_table_query)
        insert_query = f"INSERT OR REPLACE INTO '{zt_notebook.notebookId}' (id, notebook) VALUES (?, ?)"
        conn.execute(insert_query, [zt_notebook.notebookId, new_notebook])
        conn.close()
        logger.debug("Notebook with id %s loaded from toml and new db entry created", toml_data['notebookId'])

    except Exception as e:
        logger.error("Error when loading notebook, return empty notebook: %s", traceback.format_exc())

def get_notebook_request():
    global zt_notebook
    return zt_notebook

def get_notebook_db(id=''):
    conn = duckdb.connect(notebook_db_path)
    if id!="":
        get_notebook_query = f"SELECT notebook FROM '{id}' WHERE id = '{id}'"
        notebook_data = conn.execute(get_notebook_query).fetchall()
        conn.close()
    return notebook.Notebook(**json.loads(notebook_data[0][0]))

def globalStateUpdate(newCell: notebook.CodeCell=None, 
                      position_key:str=None, 
                      deletedCell: str=None, 
                      saveCell: request.SaveRequest=None, 
                      hideCell: request.HideCellRequest=None,
                      hideCode: request.HideCodeRequest=None,
                      renameCell: request.NameCellRequest=None,
                      run_request: request.Request=None, 
                      run_response: response.Response=None, 
                      new_notebook_name: str="",
                      ):
    global zt_notebook
    logger.debug("Updating state for notebook %s", zt_notebook.notebookId)
    try:        
        old_state = zt_notebook.model_dump()
        if newCell is not None:
            if position_key:
                new_cell_dict = OrderedDict()
                for k, v in zt_notebook.cells.items():
                    new_cell_dict[k] = v
                    if k==position_key:
                        new_cell_dict[newCell.id] = newCell
                zt_notebook.cells = new_cell_dict
            else:
                new_cell_dict = OrderedDict({newCell.id: newCell})
                new_cell_dict.update(zt_notebook.cells)
                zt_notebook.cells = new_cell_dict
        if deletedCell is not None:
            del zt_notebook.cells[deletedCell]
        if saveCell is not None:
            zt_notebook.cells[saveCell.id].code=saveCell.text
        if hideCell is not None:
            zt_notebook.cells[hideCell.cellId].hideCell=hideCell.hideCell
        if hideCode is not None:
            zt_notebook.cells[hideCode.cellId].hideCode=hideCode.hideCode
        if renameCell is not None:
            zt_notebook.cells[renameCell.cellId].cellName=renameCell.cellName
        if run_request is not None:
            for requestCell in run_request.cells:
                #zt_notebook.cells[requestCell.id].code = requestCell.code
                zt_notebook.cells[requestCell.id].variable_name = requestCell.variable_name
        if run_response is not None:
            for responseCell in run_response.cells:
                zt_notebook.cells[responseCell.id].components = responseCell.components
                zt_notebook.cells[responseCell.id].output = responseCell.output
                zt_notebook.cells[responseCell.id].layout = responseCell.layout
        if new_notebook_name:
            zt_notebook.notebookName = new_notebook_name
        save_notebook()
    except Exception as e:
        logger.error("Error while updating state for notebook %s: %s", zt_notebook.notebookId, traceback.format_exc())

@debounce(5)
def save_notebook():
    new_notebook = zt_notebook.model_dump_json()
    conn = duckdb.connect(notebook_db_path)
    insert_query = f"INSERT OR REPLACE INTO '{zt_notebook.notebookId}' (id, notebook) VALUES (?, ?)"
    conn.execute(insert_query, [zt_notebook.notebookId, new_notebook])
    conn.close()
    write_notebook()

def write_notebook():
    tmp_uuid_file = f'notebook_{uuid.uuid4()}.ztnb'
    logger.debug("Saving toml for notebook %s", zt_notebook.notebookId)
    try:
        with open(tmp_uuid_file, "w") as project_file:
            # Write notebookId
            project_file.write(f'notebookId = "{zt_notebook.notebookId}"\nnotebookName = "{zt_notebook.notebookName}"\n\n')

            for cell_id, cell in zt_notebook.cells.items():
                project_file.write(f'cellName = "{cell.cellName}"\n')
                project_file.write(f'[cells.{cell_id}]\n')
                project_file.write(f'cellType = "{cell.cellType}"\n')
                project_file.write(f'hideCell = "{cell.hideCell}"\n')
                project_file.write(f'hideCode = "{cell.hideCode}"\n')
                
                if cell.cellType=='sql':
                    if cell.variable_name:
                        project_file.write(f'variable_name = "{cell.variable_name}"\n')
                
                # Format code as a multi-line string
                escaped_code = cell.code.encode().decode('unicode_escape').replace('"""',"'''")
                project_file.write(f'code = """\n{escaped_code}"""\n\n')
        os.replace(tmp_uuid_file, 'notebook.ztnb')
    except Exception as e:
        logger.error("Error saving notebook %s toml file: %s", zt_notebook.notebookId, traceback.format_exc())
        
    finally:
        try:
            os.remove(tmp_uuid_file)
        except Exception as e:
            logger.debug("Error while deleting temporary toml file for notebook %s: %s", zt_notebook.notebookId, traceback.format_exc())
            pass  # Handle error silently
    logger.debug("Toml saved for notebook %s", zt_notebook.notebookId)

def get_code_completions(cell_id:str, code: str, line: int, column: int) -> list:
    try:
        script = jedi.Script(code)
        completions = script.complete(line, column)
        completions = [{"label": completion.name, "type": completion.type, "inlineFlag": False} for completion in completions]
        return {"cell_id": cell_id, "completions": completions}
    except Exception:
        logger.debug("Error getting completions for cell_id %s: %s", cell_id, traceback.format_exc())
        return {"cell_id": cell_id, "completions": []}

async def save_worker(save_queue):
    while True:
        message = await save_queue.get()
        globalStateUpdate(**message)

async def websocket_message_sender(execution_state: UserState):
    while True:
        message = await execution_state.message_queue.get()
        await execution_state.websocket.send_json(message)


def validate_color(value):
    """
    Validates the given color value.
    This validator ensures that the provided color is either one of the custom colors or follows the standard Material color palette format.
    """
    if value is None:
        return value
    custom_colors = ['primary','secondary', 'accent', 'error','info','success','warning']
    material_colors = ['red', 'pink', 'purple', 'deep-purple', 'indigo', 'blue', 'light-blue', 'cyan', 'teal', 'green', 'light-green', 'lime', 'yellow', 'amber', 'orange', 'deep-orange', 'brown', 'grey', 'blue-grey']
    if value in custom_colors:
        return value
    if '-' in value:
        color_name, shade = value.split('-')
        if color_name in material_colors and 100 <= int(shade) <= 900:
            return value
    raise ValueError(f"Invalid color value: '{value}'. Must be one of {custom_colors} or in the standard Material color palette format.")

def validate_min_less_than_max(max_value, values):
    """
    Validates that the maximum value is greater than the minimum value.
    """
    min_value = values.get('min')
    if min_value is not None and max_value <= min_value:
        raise ValueError(f"Max value '{max_value}' must be greater than min value '{min_value}'.")
    return max_value












from pydantic import BaseModel, Field, SerializeAsAny, model_validator
from typing import OrderedDict, List, Dict, Any
from uuid import uuid4
from zt_backend.models.components.zt_component import ZTComponent
from zt_backend.models.components.slider import Slider
from zt_backend.models.components.text_input import TextInput
from zt_backend.models.components.text_area_input import TextArea
from zt_backend.models.components.range_slider import RangeSlider
from zt_backend.models.components.selectbox import SelectBox
from zt_backend.models.components.multiselect import MultiSelectBox
from zt_backend.models.components.button import Button
from zt_backend.models.components.number_input import NumberInput
from zt_backend.models.components.image import Image
from zt_backend.models.components.text import Text
from zt_backend.models.components.dataframe import DataFrame
from zt_backend.models.components.layout import Layout
from zt_backend.models.components.plotly import PlotlyComponent
from zt_backend.models.components.autocomplete import Autocomplete
from zt_backend.models.components.card import Card

def deserialize_component(data: Dict[str, Any]) -> ZTComponent:
    component_map = {
        "v-slider": Slider,
        "v-text-field":TextInput,
        "v-textarea": TextArea,
        "v-number-field": NumberInput,
        "v-range-slider": RangeSlider,
        "v-select": SelectBox,
        "v-combobox": MultiSelectBox,
        "v-btn": Button,
        "v-img": Image,
        "v-text": Text,
        "v-data-table": DataFrame,
        "v-autocomplete": Autocomplete,
        "v-card": Card,
        "plotly-plot": PlotlyComponent
        # add other component types here
    }
    component_class = data.get("component")
    if component_class not in component_map:
        raise ValueError(f"Invalid component class: {component_class}")
    return component_map[component_class].model_validate(data)

class CodeCell(BaseModel):
    id: str
    code: str
    output: str
    cellName: str = Field("")
    hideCell: bool = Field(False)
    hideCode: bool = Field(False)
    variable_name: str = Field("")
    layout: Layout = Field(Layout())
    components: List[SerializeAsAny[ZTComponent]]
    cellType: str = Field(enum=['code', 'markdown', 'text', 'sql'])

    @model_validator(mode='before')
    def deserialize_components(cls, values):
        components = values.get('components', [])
        values['components'] = [deserialize_component(comp) for comp in components]
        return values
    
class Notebook(BaseModel):
    notebookName: str = Field("")
    notebookId: str = Field(default=str(uuid4()))  # Added notebook UUID
    cells: OrderedDict[str, CodeCell]
    userId: str

class Dependencies(BaseModel):
    value: str

class NotebookResponse(BaseModel):
    notebook: Notebook
    dependencies: Dependencies




from pydantic import Field
from typing import List, Optional
from zt_backend.models.components.zt_component import ZTComponent

class Card(ZTComponent):
    """A class for Card components inheriting from ZTComponent."""
    
    component: str = Field("v-card", description="Vue component name.")
    cardChildren: List[str] = Field([], description="List of child components within the card.")
    color: Optional[str] = Field(None, description="Background color of the card.")
    elevation: Optional[int] = Field(None, ge=0, le=24, description="Elevation level of the card. Must be between 0 and 24.")
    density: Optional[str] = Field(None, enum=['default','comfortable','compact'], description="Density of the component")
    width: Optional[int] = Field(None, description="Width of the card.")
    location: Optional[str] = Field(None, enum=['center'])


from pydantic import BaseModel, Field
from typing import List, Dict, Union

class CodeRequest(BaseModel):
    id: str
    code: str
    variable_name: str
    cellType: str = Field(enum=['code', 'markdown', 'text', 'sql'])

class Request(BaseModel):
    originId: str
    cells: List[CodeRequest]
    components: Dict[str, Union[str, bool, int, float, List, None]]

class Cell(BaseModel):
    code: str
    defined_names: List[str]
    loaded_names: List[str]
    loaded_modules: List[str] = []
    imported_modules: List[str] = []
    child_cells: List[int] = []
    parent_cells: List[int] = []
    previous_child_cells: List[int] = []

class CodeDict(BaseModel):
    cells: Dict[str, Cell]

class ComponentRequest(BaseModel):
    originId: str
    components: Dict[str, Union[str, bool, float, int, List, None]]
    userId: str

class DeleteRequest(BaseModel):
    cellId: str

class NotebookNameRequest(BaseModel):
    notebookName: str

class HideCellRequest(BaseModel):
    cellId: str
    hideCell: bool

class HideCodeRequest(BaseModel):
    cellId: str
    hideCode: bool

class NameCellRequest(BaseModel):
    cellId: str
    cellName: str

class CreateRequest(BaseModel):
    cellType: str = Field(enum=['code', 'markdown', 'text', 'sql'])
    position_key: str

class SaveRequest(BaseModel):
    id: str
    text: str
    code_w_context: str = Field("")
    cellType: str
    line: str = Field("")
    column: str = Field("")

class ClearRequest(BaseModel):
    userId: str

class DependencyRequest(BaseModel):
    dependencies: str


from typing import Dict, Any, OrderedDict
from io import StringIO
import pickle
from zt_backend.models import request, response
from zt_backend.runner.code_cell_parser import parse_cells, build_dependency_graph, CodeDict
from zt_backend.runner.user_state import UserState, UserContext
from zt_backend.models.components.layout import Layout
from zt_backend.utils import globalStateUpdate
from zt_backend.config import settings
from datetime import datetime
import logging
import traceback
import contextlib
import asyncio

now = datetime.now()
logger = logging.getLogger("__name__")

def try_pickle(obj):
    """
    Attempts to pickle and then unpickle an object. If successful, returns the unpickled object, 
    otherwise returns the original object.
    """
    try:
        return pickle.loads(pickle.dumps(obj))
    except Exception as e:
        logger.debug("Error during pickle: %s", e)
        return obj

def get_parent_vars(cell_id: str, code_components: CodeDict, cell_outputs_dict: Dict[str, Any]) -> Dict[str, Any]:
    """
    Gathers and returns parent variables for a given cell.

    Args:
        cell (str): ID of the current cell.
        code_components (dict): Mapping of all cell IDs to their related components such as parent cells, code, etc.
        cell_outputs_dict (dict): Contains outputs of all previously executed cells.

    Returns:
        dict: Dictionary containing the parent variables for the cell.
    """  
    if cell_id not in code_components.cells:
        # Print localstorage_data and stop execution if cell is not in code_components
        for key in cell_outputs_dict.keys():
            logger.debug("Key: %s, Value: %s", key, cell_outputs_dict[key])
        raise KeyError(f"'{cell_id}' is not present in code_components. Execution stopped.")

    parent_vars = []
    parent_cells = code_components.cells[cell_id].parent_cells
    cell_dicts = []
    for parent_cell in parent_cells:
        cell_dicts.append(cell_outputs_dict.get(parent_cell, {}))
        parent_vars_list = code_components.cells[parent_cell].loaded_names + \
                           code_components.cells[parent_cell].defined_names
        parent_vars += parent_vars_list

    exec_parents = {key: try_pickle(val) for d in cell_dicts for key, val in d.items() if key in code_components.cells[cell_id].loaded_names}

    return exec_parents

def execute_request(request: request.Request, state: UserState):
    with UserContext(state) as execution_state:
        logger.debug("Code execution started")
        cell_outputs = []
        execution_state.component_values.update(request.components)
        component_globals={'global_state': execution_state.component_values}
        dependency_graph = build_dependency_graph(parse_cells(request))
        if request.originId:
            downstream_cells = [request.originId]+dependency_graph.cells[request.originId].child_cells
            try:
                old_downstream_cells = [request.originId]
                # Find cells that are no longer dependent
                no_longer_dependent_cells = set(old_downstream_cells) - set(downstream_cells)
        
                if no_longer_dependent_cells:
                    downstream_cells.extend(list(OrderedDict.fromkeys(no_longer_dependent_cells)))
            except Exception as e:
                logger.error("Error while updating cell dependencies: %s", traceback.format_exc())
        else:
            downstream_cells = [cell.id for cell in request.cells if cell.cellType in ['code', 'sql']]
        for code_cell_id in downstream_cells:
            code_cell = dependency_graph.cells[code_cell_id]
            execution_state.message_queue.put_nowait({"cell_id": code_cell_id, "clear_output": True})
            execution_state.io_output = StringIO()
            execute_cell(code_cell_id, code_cell, component_globals, dependency_graph, execution_state)
            try:
                layout = execution_state.current_cell_layout[0]
            except Exception as e:
                logger.debug("Error while getting cell layout, setting empty layout: %s", traceback.format_exc())
                layout = Layout(**{})
            for component in execution_state.current_cell_components:
                if component.component == 'v-btn':
                    component.value = False

            cell_response = response.CellResponse(id=code_cell_id, layout=layout, components=execution_state.current_cell_components, output=execution_state.io_output.getvalue())
            cell_outputs.append(cell_response)
            execution_state.message_queue.put_nowait(cell_response.model_dump_json())
            execution_state.current_cell_components.clear()
            execution_state.current_cell_layout.clear()
        execution_state.cell_outputs_dict['previous_dependecy_graph'] = dependency_graph
        execution_state.component_values.clear()
        execution_state.created_components.clear()
        execution_state.context_globals['exec_mode'] = False
        execution_response = response.Response(cells=cell_outputs)
        execution_state.message_queue.put_nowait({"complete": True})
        if settings.run_mode=='dev':
            globalStateUpdate(run_response=execution_response,run_request=request)


def execute_cell(code_cell_id, code_cell, component_globals, dependency_graph, execution_state: UserContext):
    class WebSocketStream:
        def write(self, message):
            user_state = UserContext.get_state()
            if user_state:
                user_state.io_output.write(message)
                user_state.message_queue.put_nowait({"cell_id": code_cell_id, "output": message})

        def flush(self):
            pass
    with contextlib.redirect_stdout(WebSocketStream()), \
        contextlib.redirect_stderr(WebSocketStream()):
        try:
            if code_cell.parent_cells == []:
                temp_globals = component_globals
            else:
                temp_globals = get_parent_vars(cell_id=code_cell_id,code_components=dependency_graph,cell_outputs_dict=execution_state.cell_outputs_dict)
            execution_state.context_globals['exec_mode'] = True
            exec(code_cell.code, temp_globals)

        except Exception:
            logger.debug("Error during code execution")
            tb_list = traceback.format_exc().splitlines(keepends=True)
            tb_list = [tb_list[0]]+tb_list[3:]
            print("".join(tb_list))
    execution_state.context_globals['exec_mode'] = False
    execution_state.cell_outputs_dict[code_cell_id] = {k: try_pickle(v) for k, v in temp_globals.items() if k != '__builtins__'}


from pydantic import Field, field_validator, validator
from zt_backend.models.components.zt_component import ZTComponent
from zt_backend.models.validations import validate_color
from typing import Optional, Union
from zt_backend.runner.user_state import UserContext

class Slider(ZTComponent):
    """A slider component that allows you to capture numeric input from a user. 
    
    To use the slider simply import the package and then declare the slider. The only required 
    field is an id. Your slider will render with default max and min values and a number 
    of other defaults.
    
    """
    component: str = Field("v-slider", description="Vue component name.")
    value: Union[int,float] = Field(0, description="Current value of the slider.")
    min: Union[int,float] = Field(0,  description="Minimum value of the slider.")
    max: Union[int,float] = Field(100, description="Maximum value of the slider.")
    step: Union[int,float] = Field(1,  description="Step increment of the slider.")
    thumb_label: bool = Field(False, description="Displays the thumb label.")
    thumb_size: int = Field(0, description="Size of the thumb.")
    tick_labels: str = Field('always', description="Displays the tick labels.")
    ticks: list = Field([], description="Displays the ticks.")
    color: str = Field('primary', pre=True, description="Color of the range slider. Can be custom or standard Material color.")
    size: str = Field('large', description="Size of the slider.")
    label: Optional[str] = Field(None,description= 'A label for your slider')
    rounded: bool = Field(True, description="Determines if the slider has rounded edges.")
    triggerEvent: str = Field('end',description="Trigger event for when to run the slider")
    
    @field_validator('color')
    def validate_color(cls, color):
        return validate_color(color)
    
    @validator('value', always=True) #TODO: debug and replace with field validator
    def get_value_from_global_state(cls, value, values):
        id = values['id'] # Get the id if it exists in the field values
        execution_state = UserContext.get_state()
        try:
            if execution_state and id and id in execution_state.component_values:  # Check if id exists in global_state
                return execution_state.component_values[id]  # Return the value associated with id in global_state
        except Exception as e:
            e
        return value  # If id doesn't exist in global_state, return the original value


from pydantic import Field, validator
from zt_backend.models.components.zt_component import ZTComponent
from typing import Union,Optional
from zt_backend.runner.user_state import UserContext

class NumberInput(ZTComponent):
    """A class for TextInput components inheriting from ZTComponent."""
    component: str = Field("v-number-field", description="Vue component name.")
    value: Union[int,float,None] = Field (None,description="The input text value")
    placeholder: Optional[str] = Field(None, description="Placeholder text.")
    label: Optional[str] = Field(None, description="Label for the text input.")
    readonly: Optional[bool] = Field(None, description="If true, the input is read-only.")
    disabled: Optional[bool] = Field(None, description="If true, the input is disabled.")
    type: str = Field('number',description="Ensures that only numbers are accepted on the frontend")
    triggerEvent: str = Field('input',description="Trigger event to send code to the backend")
    
    @validator('value', always=True) #TODO: debug and replace with field validator
    def get_value_from_global_state(cls, value, values):
        if value=="":
            return None
        id = values['id'] # Get the id if it exists in the field values
        execution_state = UserContext.get_state()
        try:
            if execution_state and id and id in execution_state.component_values:  # Check if id exists in global_state
                return execution_state.component_values[id]  # Return the value associated with id in global_state
        except Exception as e:
            e
        return value  # If id doesn't exist in global_state, return the original value


import subprocess
from fastapi import APIRouter, WebSocket, WebSocketDisconnect
from zt_backend.models import request, notebook
from zt_backend.runner.execute_code import execute_request
from zt_backend.config import settings
from zt_backend.utils import *
from zt_backend.runner.user_state import UserState
from fastapi.responses import HTMLResponse
from copilot.copilot import text_document_did_change
import logging
import site
import uuid
import os
import threading
import traceback
import sys
import asyncio
import pkg_resources
import trace

class ConnectionManager:
    def __init__(self):
        self.active_connections: list[WebSocket] = []

    async def connect(self, websocket: WebSocket):
        await websocket.accept()
        self.active_connections.append(websocket)

    def disconnect(self, websocket: WebSocket):
        self.active_connections.remove(websocket)

    async def send_text(self, message: str, websocket: WebSocket):
        await websocket.send_text(message)

    async def broadcast(self, message: str):
        for connection in self.active_connections:
            await connection.send_text(message)

class KThread(threading.Thread):
  """A subclass of threading.Thread, with a kill()
method."""
  def __init__(self, *args, **keywords):
    threading.Thread.__init__(self, *args, **keywords)
    self.killed = False

  def start(self):
    """Start the thread."""
    self.__run_backup = self.run
    self.run = self.__run      # Force the Thread to install our trace.
    threading.Thread.start(self)

  def __run(self):
    """Hacked run function, which installs the
trace."""
    sys.settrace(self.globaltrace)
    self.__run_backup()
    self.run = self.__run_backup

  def globaltrace(self, frame, why, arg):
    if why == 'call':
      return self.localtrace
    else:
      return None

  def localtrace(self, frame, why, arg):
    if self.killed:
      if why == 'line':
        raise SystemExit()
    return self.localtrace

  def kill(self):
    self.killed = True

router = APIRouter()
manager = ConnectionManager()
current_path = os.path.dirname(os.path.abspath(__file__))

#connect to db for saving notebook
notebook_db_dir =  site.USER_SITE+'/.zero_true/'
notebook_db_path = notebook_db_dir+'notebook.db'
os.makedirs(notebook_db_dir, exist_ok=True)

user_states={}
user_timers={}
user_threads={}
user_message_tasks={}
notebook_state=UserState('')
run_mode = settings.run_mode
save_queue = asyncio.Queue()

logger = logging.getLogger("__name__")

@router.get("/app", response_class=HTMLResponse)
async def catch_all():
    if(run_mode=='dev'):
        return HTMLResponse(open(os.path.join(current_path, "dist_dev", "index.html")).read())

@router.get("/health")
def health():
    return('UP')

@router.get("/env_data")
def env_data():
    return {
        "ws_url": settings.ws_url,
        "python_version": f'{sys.version_info.major}.{sys.version_info.minor}',
        "zt_version": pkg_resources.get_distribution('zero-true').version
    }

@router.get("/base_path")
def base_path():
    return settings.user_name + '/' + settings.project_name

@router.websocket("/ws/run_code")
async def run_code(websocket: WebSocket):
    global current_thread
    if(run_mode=='dev'):
        message_send = asyncio.create_task(websocket_message_sender(notebook_state))
        await manager.connect(websocket)
        try:
            while True:
                data = await websocket.receive_json()
                ws_request = request.Request(**data)
                notebook_state.websocket = websocket
                current_thread = KThread(target = execute_request, args=(ws_request, notebook_state))
                current_thread.start()
        except WebSocketDisconnect:
            manager.disconnect(websocket)
        finally:
            message_send.cancel() 

@router.websocket("/ws/component_run")
async def component_run(websocket: WebSocket):
    global current_thread
    global user_threads
    await manager.connect(websocket)
    try:
        while True:
            data = await websocket.receive_json()
            logger.debug("Component change code execution started")
            notebook = get_notebook_request()
            component_request = request.ComponentRequest(**data)
            cells = []
            for cell_key, cell in notebook.cells.items():
                cell_request=request.CodeRequest(
                    id=cell.id, 
                    code=cell.code,
                    variable_name=cell.variable_name,
                    cellType=cell.cellType
                )
                cells.append(cell_request)
            code_request = request.Request(
                originId=component_request.originId,
                cells=cells,
                components=component_request.components
            )
            if(run_mode=='dev'):
                notebook_state.websocket = websocket
                current_thread = KThread(target = execute_request, args=(code_request, notebook_state))
                current_thread.start()
            else:
                if component_request.userId not in user_states:
                    logger.debug("New user execution with id: %s, sending refresh", component_request.userId)
                    await websocket.send_json({"refresh": True})
                logger.debug("Existing user execution with id: %s", component_request.userId)
                timer_set(component_request.userId, 1800)
                user_states[component_request.userId].websocket = websocket
                user_threads[component_request.userId] = KThread(target = execute_request, args=(code_request, user_states[component_request.userId]))
                user_threads[component_request.userId].start()
    except WebSocketDisconnect:
        manager.disconnect(websocket)

@router.post("/api/create_cell")
def create_cell(cellRequest: request.CreateRequest):
     if(run_mode=='dev'):
        logger.debug("Code cell addition request started")
        createdCell = notebook.CodeCell(
            id=str(uuid.uuid4()),
            code='',
            components=[],
            output='',
            variable_name='',
            cellType=cellRequest.cellType
        )
        save_queue.put_nowait({"newCell": createdCell.model_copy(deep=True), "position_key":cellRequest.position_key})
        logger.debug("Code cell addition request completed")
        return createdCell

@router.post("/api/delete_cell")
def delete_cell(deleteRequest: request.DeleteRequest):
     cell_id = deleteRequest.cellId
     if(run_mode=='dev'):
        try:
            notebook_state.cell_outputs_dict.pop(cell_id, None)
        except Exception as e:
            logger.error("Error when deleting cell %s from cell_outputs_dict: %s", cell_id, traceback.format_exc())
        try:
            cell_dict = notebook_state.cell_outputs_dict['previous_dependecy_graph'].cells
            if cell_id in cell_dict:
                cell_dict.pop(cell_id, None)

            # Recursively search for and remove the cell ID from child_cells and parent_cells in other cells
            for cell_key, cell in cell_dict.items():
                if cell_id in dict(cell).get("child_cells", []):
                    cell["child_cells"].pop(cell_id, None)
                if cell_id in dict(cell).get("parent_cells", []):
                    cell["parent_cells"].pop(cell_id, None)

        except Exception as e:
            logger.debug("Error when deleting cell %s from cell dicts: %s", cell_id, traceback.format_exc())
        logger.debug("Cell %s deleted successfully", cell_id)
        save_queue.put_nowait({"deletedCell":cell_id})

@router.post("/api/hide_cell")
def hide_cell(hideCellRequest: request.HideCellRequest):
     if(run_mode=='dev'):
        logger.debug("Hide cell request started")
        save_queue.put_nowait({"hideCell": hideCellRequest})
        logger.debug("Hide cell request completed")

@router.post("/api/hide_code")
def hide_cell(hideCodeRequest: request.HideCodeRequest):
     if(run_mode=='dev'):
        logger.debug("Hide code request started")
        save_queue.put_nowait({"hideCode": hideCodeRequest})
        logger.debug("Hide code request completed")

@router.post("/api/rename_cell")
def hide_cell(renameCellRequest: request.NameCellRequest):
     if(run_mode=='dev'):
        logger.debug("Rename cell request started")
        save_queue.put_nowait({"renameCell": renameCellRequest})
        logger.debug("Rename cell request completed")

@router.websocket("/ws/save_text")
async def save_text(websocket: WebSocket):
    if(run_mode=='dev'):
        save_task = asyncio.create_task(save_worker(save_queue))
        await manager.connect(websocket)
        try:
            while True:
                data = await websocket.receive_json()
                cell_type = data.get("cellType")
                code = data.get("text")
                cell_id = data.get("id")
                save_request = request.SaveRequest(id=cell_id, text=code, cellType=cell_type)
                save_queue.put_nowait({"saveCell":save_request})
                if cell_type=="code":
                    line = data.get("line")
                    column = data.get("column")
                    await text_document_did_change({
                        "textDocument": {
                            "uri": "file:///notebook.ztnb",
                            "version": 1
                        },
                        "contentChanges": [{
                            "text": data.get("code_w_context")
                        }]
                    })
                    code_w_context= data.get("code_w_context")
                    completions = get_code_completions(cell_id, code_w_context, line, column)
                    await websocket.send_json(completions)
        except WebSocketDisconnect:
            manager.disconnect(websocket)
        finally:
            save_task.cancel()

@router.post("/api/clear_state")
def clear_state(clearRequest: request.ClearRequest):
     if(run_mode=='app'):
        logger.debug("Clearing state for user %s", clearRequest.userId)
        user_states.pop(clearRequest.userId, None)

@router.post("/api/dependency_update")
def dependency_update(dependencyRequest: request.DependencyRequest):
     if(run_mode=='dev'):
        logger.debug("Updating dependencies")
        try:
            with open('requirements.txt', 'r+', encoding='utf-8') as file:
                contents = file.read()
                if contents == dependencyRequest.dependencies:
                    return "No change to dependencies"
                file.seek(0)
                file.write(dependencyRequest.dependencies)
                file.truncate()
                subprocess.run(['pip', 'install', '-r', 'requirements.txt'])
                subprocess.run(['lock', 'requirements.txt'])
                logger.debug("Successfully updated dependencies")
        except Exception as e:
            logger.error('Error while updating requirements: %s', traceback.format_exc())

@router.websocket("/ws/notebook")
async def load_notebook(websocket: WebSocket):
    global user_threads
    await manager.connect(websocket)
    try:
        while True:
            await websocket.receive_text()
            logger.debug("Get notebook request received")
            notebook_start = get_notebook_request()
            await websocket.send_json({"notebook_name": notebook_start.notebookName})
            if (run_mode=='app'):
                userId = str(uuid.uuid4())
                notebook_start.userId = userId
                user_states[userId]=UserState(userId)
                user_message_tasks[userId]=asyncio.create_task(websocket_message_sender(user_states[userId]))
                timer_set(userId, 1800)
                cells = []
                components={}

                for cell_key, cell in notebook_start.cells.items():
                    cell_request=request.CodeRequest(
                        id=cell.id, 
                        code=cell.code,
                        variable_name=cell.variable_name,
                        cellType=cell.cellType
                    )
                    for comp in cell.components:
                        if hasattr(comp, 'value'):
                            components[comp.id] = comp.value
                    cells.append(cell_request)
                code_request = request.Request(
                    originId='',
                    cells=cells,
                    components=components
                )
                notebook_response = notebook.NotebookResponse(notebook=notebook_start, dependencies=notebook.Dependencies(value=''))
                await websocket.send_json(notebook_response.model_dump_json())
                user_states[userId].websocket = websocket
                user_threads[userId] = KThread(target = execute_request, args=(code_request, user_states[userId]))
                user_threads[userId].start()
            else:
                try:
                    with open('requirements.txt', 'r', encoding='utf-8') as file:
                        contents = file.read()
                    notebook_response =  notebook.NotebookResponse(notebook=notebook_start, dependencies=notebook.Dependencies(value=contents))
                    await websocket.send_json(notebook_response.model_dump_json())
                    await websocket.send_json({"complete": True})
                except FileNotFoundError:
                    logger.error('Requirements file not found')

    except WebSocketDisconnect:
        manager.disconnect(websocket)

@router.websocket("/ws/stop_execution")
async def stop_execution(websocket: WebSocket):
    global current_thread
    global user_threads
    await manager.connect(websocket)
    try:
        while True:
            data = await websocket.receive_text()
            if run_mode=='dev' and current_thread:
                current_thread.kill()
                notebook_state.current_cell_components.clear()
                notebook_state.current_cell_layout.clear()
                notebook_state.component_values.clear()
                notebook_state.created_components.clear()
                notebook_state.context_globals['exec_mode'] = False
            if run_mode=='app' and user_threads[data]:
                user_threads[data].kill()
                user_states[data].current_cell_components.clear()
                user_states[data].current_cell_layout.clear()
                user_states[data].component_values.clear()
                user_states[data].created_components.clear()
                user_states[data].context_globals['exec_mode'] = False
    except WebSocketDisconnect:
        manager.disconnect(websocket)

@router.post("/api/notebook_name_update")
def notebook_name_update(notebook_name: request.NotebookNameRequest):
    if(run_mode=='dev'):
        save_queue.put_nowait({"new_notebook_name":notebook_name.notebookName})

@router.on_event('shutdown')
def shutdown():
    global current_thread
    global user_threads
    if current_thread:
        current_thread.kill()
    for user_id in user_threads:
        if user_threads[user_id]:
            user_threads[user_id].kill()
    for user_id in user_timers:
        if user_timers[user_id]:
            user_timers[user_id].cancel()
    for user_id in user_message_tasks:
        if user_message_tasks[user_id]:
            user_message_tasks[user_id].cancel()

def remove_user_state(user_id):
    try:
        if user_id in user_timers:
            # Cancel and remove the associated timer
            timer = user_timers[user_id]
            message_sender = user_message_tasks[user_id]
            if timer:
                timer.cancel()
            del user_timers[user_id]
            if message_sender:
                message_sender.cancel() 
            del user_message_tasks[user_id]
            if user_id in user_states: del user_states[user_id]
            logger.debug("User state removed for user %s", user_id)
    except Exception as e:
        logger.error("Error removing user state for user %s: %s", user_id, traceback.format_exc())

def timer_set(user_id, timeout_seconds):
    logger.debug("Starting timer for user %s", user_id)
    if user_id in user_timers:
        existing_timer = user_timers[user_id]
        if existing_timer:
            existing_timer.cancel()
        
        timer = threading.Timer(timeout_seconds, remove_user_state, args=(user_id,))
        timer.daemon=True
        timer.start()
        
        user_timers[user_id] = timer



#!/usr/bin/env python3

import os
import typer
from typing_extensions import Annotated
import uvicorn
from typing import Optional
from rich import print
import shutil
import requests
import pkg_resources
import yaml

cli_app = typer.Typer()

def print_ascii_logo():
    ascii_logo="""

_______________________________ ________    _____________________ ____ ______________
\____    /\_   _____/\______   \\_____  \   \__    ___/\______   \    |   \_   _____/
  /     /  |    __)_  |       _/ /   |   \    |    |    |       _/    |   /|    __)_ 
 /     /_  |        \ |    |   \/    |    \   |    |    |    |   \    |  / |         |
/_______ \/_______  / |____|_  /\_______  /   |____|    |____|_  /______/ /_______  /
        \/        \/         \/         \/                     \/                 \/ 

    """
    print(f"[purple]{ascii_logo}[/purple]")

def upload_to_s3(signed_url, zip_path):
    try:
        files = { 'file': open(zip_path, 'rb')}
        r = requests.post(signed_url['url'],data=signed_url['fields'],files=files)
        if r.status_code == 204:
            typer.echo("Successfully uploaded to S3.")
        else:
            typer.echo(f"Failed to upload to S3: {r.content}")
    except Exception as e:
        typer.echo(f"An error occurred: {e}")

@cli_app.command()
def publish(key: Annotated[Optional[str],typer.Argument(help="The API key used to publish your project")],
            user_name: Annotated[Optional[str],typer.Argument(help="User Name to publish the project under")],
            project_name:  Annotated[Optional[str],typer.Argument(help="The project to be published")],
            project_source: Annotated[Optional[str],typer.Argument(help="The directory for your project")]):
    
    s3_key = user_name+'/'+ project_name+'/'+project_name+'.zip'
    # Step 1: Verify the API key and get a signed URL from the Lambda function
    lambda_url = "https://mxj5v1635d.execute-api.us-east-1.amazonaws.com/default/project_upload"
    headers = {"Content-Type":"application/json",
               "x-api-key":key}
    response = requests.post(lambda_url, json={"s3_key": s3_key}, headers=headers)

    if response.status_code != 200:
        typer.echo(f"Execution error occured: {response.content}")
        return
    signed_url = response.json().get("uploadURL")
    if not signed_url:
        typer.echo("Failed to get a signed URL.")
        return

    # Step 2: Zip the project files using shutil
    output_filename = f"{project_name}"
    if project_source == '.':
        project_source = os.path.basename(os.path.normpath(os.getcwd()))
        print(project_source)
        os.chdir('..')

    shutil.make_archive(base_name =output_filename,format='zip', root_dir = project_source)

    # Step 3: Upload zip to S3 using the signed URL
    upload_to_s3(signed_url, f"{output_filename}.zip")
    os.remove(f"{output_filename}.zip")
    return signed_url

@cli_app.command()
def app(host: Annotated[Optional[str], typer.Argument(help="Host address to bind to.")]="localhost",
        port: Annotated[Optional[int], typer.Argument(help="Port number to bind to.")]=2613):
    """
    Start the Zero-True application.
    """

    print_ascii_logo()
    os.environ['RUN_MODE'] = 'app'
    log_path = os.path.normpath(pkg_resources.resource_filename('zt_cli', 'log_config.yaml'))
        
    print(f"[yellow]Starting Zero-True in app mode on http://{host}:{port}[/yellow]")

    with open(log_path) as f:
        log_config_dict = yaml.full_load(f)
    
    uvicorn.run('zt_backend.main:app', host=host, port=port, log_config=log_config_dict)

@cli_app.command()
def notebook(host: Annotated[Optional[str], typer.Argument(help="Host address to bind to.")]="localhost",
        port: Annotated[Optional[int], typer.Argument(help="Port number to bind to.")]=1326):
    """
    Start the Zero-True application.
    """

    print_ascii_logo()
    os.environ['RUN_MODE'] = 'dev'
    log_path = os.path.normpath(pkg_resources.resource_filename('zt_cli', 'log_config.yaml'))
        
    print(f"[yellow]Starting Zero-True in notebook mode on http://{host}:{port}[/yellow]")

    with open(log_path) as f:
        log_config_dict = yaml.full_load(f)
    
    uvicorn.run('zt_backend.main:app', host=host, port=port, log_config=log_config_dict)

if __name__ == "__main__":
    cli_app()


from pydantic import BaseModel, SerializeAsAny, Field
from typing import List
from zt_backend.models.components.zt_component import ZTComponent
from zt_backend.models.components.layout import Layout
from typing import Union


class CellResponse(BaseModel):
    id: str
    components: List[SerializeAsAny[ZTComponent]]
    layout: Union[Layout,None]
    output: str

class Response(BaseModel):
    cells: List[CellResponse]
    refresh: bool = Field(False)


import threading
import asyncio

class UserState:
    def __init__(self, user_id):
        self.user_id = user_id
        self.component_values = {}
        self.created_components = []
        self.context_globals = {"exec_mode": False}
        self.current_cell_components = []
        self.current_cell_layout = []
        self.cell_outputs_dict = {}
        self.websocket = None
        self.io_output = None
        self.message_queue = asyncio.Queue()

class UserContext:
    _state = threading.local()

    @staticmethod
    def set_state(state: UserState):
        UserContext._state.value = state

    @staticmethod
    def get_state():
        return getattr(UserContext._state, 'value', None)

    def __init__(self, user_state: UserState):
        self.user_state = user_state

    def __enter__(self):
        UserContext.set_state(self.user_state)
        return self.user_state

    def __exit__(self, exc_type, exc_val, exc_tb):
        UserContext.set_state(None)


from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from fastapi.middleware.cors import CORSMiddleware
from zt_backend.config import settings
from zt_backend.utils import get_notebook, write_notebook
from copilot.copilot import copilot_app
import zt_backend.router as router
import os
import subprocess
import logging
import traceback

app = FastAPI()
logger = logging.getLogger("__name__")

current_path = os.path.dirname(os.path.abspath(__file__))

run_mode = settings.run_mode
project_name = settings.project_name
user_name = settings.user_name

route_prefix = ''
if project_name:
    route_prefix = '/'+user_name+'/'+project_name

app.include_router(router.router, prefix=route_prefix)
app.mount(route_prefix+"/copilot", copilot_app)

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=False,
    allow_methods=["*"],
    allow_headers=["*"],
    expose_headers=["*"]
)

@app.on_event("startup")
def open_project():
    try:
        if not os.path.exists('notebook.ztnb'):
            logger.info("No toml file found, creating with empty notebook")
            write_notebook()
        if not os.path.exists('requirements.txt'):
            logger.info("No requirements file found, creating with base dependency")
            with open('requirements.txt', 'w') as file:
                file.write('zero-true')
            try:
                subprocess.run(['lock', 'requirements.txt'])
            except Exception:
                logger.error("Failed to lock requirements: %s", traceback.format_exc())
        get_notebook()
    except Exception as e:
        logger.error("Error creating new files on startup: %s", traceback.format_exc())
        
if run_mode=='app':
    app.mount(route_prefix, StaticFiles(directory=os.path.join(current_path, "dist_app"), html=True), name="assets")
else:
    app.mount(route_prefix, StaticFiles(directory=os.path.join(current_path, "dist_dev"), html=True), name="assets")


from pydantic import BaseModel, Field
from typing import List
from typing_extensions import Literal

class BlankRequest(BaseModel):
    # Add fields if the signInInitiate request needs any parameters
    pass

class SignInConfirmRequest(BaseModel):
    userCode: str  # Assuming signInConfirm requires a userCode field

class TextDocumentItem(BaseModel):
    uri: str
    languageId: str
    version: int
    text: str

class DidOpenTextDocumentParams(BaseModel):
    textDocument: TextDocumentItem

class Position(BaseModel):
    line: int
    character: int

class DocumentItem(BaseModel):
    version: int
    position: Position
    uri: str

class GetCompletionsParams(BaseModel):
    doc: DocumentItem


class VersionedTextDocumentIdentifier(BaseModel):
    uri: str
    version: int

class TextDocumentContentChangeEvent(BaseModel):
    text: str  # The new text of the whole document

class DidChangeTextDocumentParams(BaseModel):
    textDocument: VersionedTextDocumentIdentifier
    contentChanges: List[TextDocumentContentChangeEvent] = Field(default_factory=list)


class CopilotPayloadSignInInitiate(BaseModel):
    verificationUri: str
    status: str
    userCode: str
    expiresIn: int
    interval: int

class CopilotPayloadSignInConfirm(BaseModel):
    status: Literal["AlreadySignedIn", "MaybeOk", "NotAuthorized", "NotSignedIn", "OK"]
    user: str

class CopilotPayloadSignOut(BaseModel):
    status: Literal["NotSignedIn"]


class Position(BaseModel):
    line: int
    character: int

class Range(BaseModel):
    start: Position
    end: Position

class Completion(BaseModel):
    text: str
    position: Position
    uuid: str
    range: Range
    displayText: str

class CopilotGetCompletionsResult(BaseModel):
    completions: List[Completion]

class AcceptRequest(BaseModel):
    uuid: str

class RejectRequest(BaseModel):
    uuid: str


from pydantic import Field, validator
from zt_backend.models.components.zt_component import ZTComponent
from zt_backend.runner.user_state import UserContext

class Button(ZTComponent):
    """A class for Button components inheriting from ZTComponent."""
    component: str = Field("v-btn", description="Vue component name.")
    value: bool = Field (False, description="Whether the button has been clicked")
    text: str = Field("Click Me", description="Label displayed on the button.")
    color: str = Field("primary", description="Color of the button.")
    disabled: bool = Field(False, description="If true, the button is disabled.")
    outlined: bool = Field(False, description="If true, the button will have an outlined style.")
    triggerEvent: str = Field("click", description="Trigger event to send code to the backend")

    @validator('value', always=True)
    def get_label_from_global_state(cls, value, values):
        id = values.get('id')  # Get the id if it exists in the field values
        execution_state = UserContext.get_state()
        try:
            if execution_state and id and id in execution_state.component_values:  # Check if id exists in global_state
                return execution_state.component_values[id]  # Return the value associated with id in global_state
        except Exception as e:
            pass  # Handle exception as needed
        return (value)  # If id doesn't exist in global_state, return the original value



from pydantic_settings import BaseSettings
from functools import lru_cache

class Settings(BaseSettings):
    run_mode: str = "app"
    project_name: str = ""
    user_name: str = ""
    ws_url: str = ""

@lru_cache()
def get_settings():
    return Settings()

settings = get_settings()


from pydantic import Field
from zt_backend.models.components.zt_component import ZTComponent
import matplotlib.pyplot as plt
import base64
from io import BytesIO

class Matplotlib(ZTComponent):
    component: str = Field("v-img", description="Vue component name.")
    src: str = Field(..., description="Source URL of the image.")
    alt: str = Field("", description="Alternative text for the image.")
    width: int = Field(100, description="Width of the image.")
    height: int = Field(100, description="Height of the image.")
    
    @classmethod
    def from_matplotlib(cls,id: str, figure: plt.Figure, alt=None, width=None, height=None):
        plt.style.use('dark_background')
        buffer = BytesIO()
        figure.savefig(buffer, format="png")
        buffer.seek(0)
        base64_data = base64.b64encode(buffer.read()).decode()
        src = f"data:image/png;base64,{base64_data}"
        return cls(id=id,src=src, alt=alt, width=width, height=height)


from .request import DependencyRequest, Request, ComponentRequest, DeleteRequest, CreateRequest, SaveRequest, ClearRequest, NotebookNameRequest, HideCellRequest, HideCodeRequest, NameCellRequest
from .response import Response
from .notebook import Notebook
from .notebook import NotebookResponse
from .components.slider import Slider
from .components.text_input import TextInput
from .components.text_area_input import TextArea
from .components.range_slider import RangeSlider
from .components.selectbox import SelectBox
from .components.multiselect import MultiSelectBox
from .components.button import Button
from .components.number_input import NumberInput
from .components.image import Image
from .components.text import Text
from .components.layout import Layout
from .components.autocomplete import Autocomplete
from .components.card import Card
import json

def generate_json(model, name):
    with open('zt_schema/'+name+'.json', 'w+') as file:
        file.write(json.dumps(model.model_json_schema(),indent=2))

def generate_schema():
    generate_json(Request, 'request')
    generate_json(ComponentRequest, 'component_request')
    generate_json(DeleteRequest, 'delete_request')
    generate_json(CreateRequest, 'create_request')
    generate_json(SaveRequest, 'save_request')
    generate_json(ClearRequest, 'clear_request')
    generate_json(DependencyRequest, 'dependency_request')
    generate_json(NotebookNameRequest, 'notebook_name_request')
    generate_json(HideCellRequest, 'hide_cell_request')
    generate_json(HideCodeRequest, 'hide_code_request')
    generate_json(NameCellRequest, 'name_cell_request')
    generate_json(Response, 'response')
    generate_json(Slider, 'slider')
    generate_json(Notebook, 'notebook')
    generate_json(NotebookResponse, 'notebook_response')
    generate_json(TextInput, 'text_input')
    generate_json(TextArea,'text_area')
    generate_json(RangeSlider,'range_slider')
    generate_json(SelectBox,'select_box')
    generate_json(MultiSelectBox,'multiselect_box')
    generate_json(Text,'text')
    generate_json(Button,'button')
    generate_json(NumberInput,'number_input')
    generate_json(Image,'image')
    generate_json(Layout,'layout')
    generate_json(Autocomplete,'autocomplete')
    generate_json(Card,'card')
    generate_json(Text,'text')


from typing import Optional
from pydantic import Field
from zt_backend.models.components.zt_component import ZTComponent

class Text(ZTComponent):
    """A class for static text components inheriting from ZTComponent."""
    component: str = Field("v-text", description="Vue component name for static text.")
    text: str = Field("", description="The actual text content.")
    type: str = Field('text-body-1', enum=[
        'text-h1', 'text-h2', 'text-h3', 'text-h4', 'text-h5', 'text-h6',
        'text-subtitle-1', 'text-subtitle-2', 'text-body-1', 'text-body-2',
        'text-button', 'text-caption', 'text-overline'
    ], description="Text type for text-related components.")
    color: Optional[str] = Field(None, description="Color of the text.")


#!/usr/bin/env python3

import subprocess
import os
import shutil
import typer
from zt_backend.models.generate_schema import generate_schema
from typing_extensions import Annotated
from typing import Optional
from rich import print
import pkg_resources

cli_app = typer.Typer()

def print_ascii_logo():
    ascii_logo="""
_____________________ ________  _______________   ____
\____    /\__    ___/ \______ \ \_   _____/\   \ /   /
  /     /   |    |     |    |  \ |    __)_  \   Y   / 
 /     /_   |    |     |    `   \|        \  \     /  
/_______ \  |____|    /_______  /_______  /   \___/   
        \/                    \/        \/            

    """
    print(f"[purple]{ascii_logo}[/purple]")

def generate_ts():
    os.mkdir('zt_schema')
    generate_schema()
    os.chdir('zt_frontend')
    shutil.rmtree('src/types')
    os.system('yarn json2ts -i ../zt_schema -o src/types')
    os.chdir('..')
    shutil.rmtree('zt_schema')

def build_frontend():
    try:
        shutil.rmtree('zt_backend/dist_dev')
        shutil.rmtree('zt_backend/dist_app')
    except Exception as e:
        typer.echo(e)
    os.chdir('zt_frontend')
    os.system('yarn install')
    os.system('yarn run build')
    os.chdir('..')
    shutil.copytree('zt_frontend/dist', 'zt_backend/dist_dev')
    os.chdir('zt_frontend')
    os.system('yarn run buildapp')
    os.chdir('..')
    shutil.copytree('zt_frontend/dist', 'zt_backend/dist_app')

@cli_app.command()
def pyd2ts():
    generate_ts()

@cli_app.command()
def build():
    build_frontend()    

@cli_app.command()
def app(port: Annotated[Optional[int], typer.Argument(help="Port number to bind to.")]=5173):
    
    print_ascii_logo()

    log_path = os.path.normpath(pkg_resources.resource_filename('zt_dev_cli', 'log_config.yaml'))
    os.environ['RUN_MODE'] = 'app'
    frontend_cmd = ["yarn", "run", "app", str(port)]
    backend_cmd = ["uvicorn", "zt_backend.main:app", "--reload", f"--log-config={log_path}"]
    if (os.name == 'nt'):
        backend_cmd = ["start"]+backend_cmd

    backend_process = subprocess.Popen(backend_cmd, shell=(os.name == 'nt'))
    os.chdir("zt_frontend")
    frontend_process = subprocess.Popen(frontend_cmd, shell=(os.name == 'nt'))

    backend_process.wait()
    frontend_process.wait()

@cli_app.command()
def notebook(port: Annotated[Optional[int], typer.Argument(help="Port number to bind to.")]=5173):
    
    print_ascii_logo()

    log_path = os.path.normpath(pkg_resources.resource_filename('zt_dev_cli', 'log_config.yaml'))
    os.environ['RUN_MODE'] = 'dev'
    frontend_cmd = ["yarn", "run", "dev", str(port)]
    backend_cmd = ["uvicorn", "zt_backend.main:app", "--reload", f"--log-config={log_path}"]
    if (os.name == 'nt'):
        backend_cmd = ["start"]+backend_cmd

    backend_process = subprocess.Popen(backend_cmd, shell=(os.name == 'nt'))
    os.chdir("zt_frontend")
    frontend_process = subprocess.Popen(frontend_cmd, shell=(os.name == 'nt'))

    backend_process.wait()
    frontend_process.wait()

if __name__ == "__main__":
    cli_app()


from fastapi import FastAPI, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from copilot.copilot_models import BlankRequest, DidOpenTextDocumentParams,\
                           SignInConfirmRequest, GetCompletionsParams\
                            ,DidChangeTextDocumentParams, CopilotPayloadSignInInitiate\
                            ,CopilotPayloadSignInConfirm, CopilotGetCompletionsResult\
                            ,CopilotPayloadSignOut, TextDocumentItem, AcceptRequest, RejectRequest
import requests
from typing import Union
import os
from threading import Timer
import asyncio

copilot_app = FastAPI()

copilot_app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=False,
    allow_methods=["*"],
    allow_headers=["*"],
    expose_headers=["*"]
)

copilot_enabled = False
copilot_doc_open = False
debounce_timer = None
version = 0

NODE_SERVER_URL = "http://localhost:3000"
NODE_SERVER_SCRIPT_PATH = os.path.dirname(os.path.abspath(__file__))+"/client.js"
NODE_SERVER_NAME = "node-server"

async def start_node_server():
    if not os.path.isfile(NODE_SERVER_SCRIPT_PATH):
        raise Exception(f"Node.js script not found at {NODE_SERVER_SCRIPT_PATH}")

    try:
        proc = await asyncio.create_subprocess_shell(
            "node "+NODE_SERVER_SCRIPT_PATH,
            stdin=asyncio.subprocess.DEVNULL,
            stdout=asyncio.subprocess.DEVNULL,
            stderr=asyncio.subprocess.DEVNULL)
        print("Node server started")
    except Exception as e:
        print(f"An error occurred while starting server: {e}")
        raise HTTPException(status_code=500, detail=str(e))

@copilot_app.post("/start_node_server")
async def start_node_server_route():
    try:
        await start_node_server()
        return {"message": 'Node server started successfully'}
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))

@copilot_app.post("/check_status",response_model=Union[CopilotPayloadSignInConfirm,CopilotPayloadSignOut])
async def check_status(req: BlankRequest):
    global copilot_enabled
    try:
        response = requests.post(f"{NODE_SERVER_URL}/sendRequest", json={"method": "checkStatus", "params": {}})
        response.raise_for_status()
        status = response.json().get('status', '')
        if status == "OK" or status == "AlreadySignedIn":
            copilot_enabled = True
        return response.json()
    except requests.RequestException as e:
        raise HTTPException(status_code=500, detail=str(e))

@copilot_app.post("/sign_in_initiate", response_model=Union[CopilotPayloadSignInInitiate, CopilotPayloadSignInConfirm])
async def sign_in_initiate(req: BlankRequest):
    global copilot_enabled
    try:
        response = requests.post(f"{NODE_SERVER_URL}/sendRequest", json={"method": "signInInitiate", "params": req.dict()})
        response.raise_for_status()
        status = response.json().get('status', '')
        if status == "OK" or status == "AlreadySignedIn":
            copilot_enabled = True
        return response.json()
    except requests.RequestException as e:
        raise HTTPException(status_code=500, detail=str(e))

@copilot_app.post("/sign_in_confirm",response_model=CopilotPayloadSignInConfirm)
async def sign_in_confirm(req: SignInConfirmRequest):
    global copilot_enabled
    try:
        response = requests.post(f"{NODE_SERVER_URL}/sendRequest", json={"method": "signInConfirm", "params": req.dict()})
        response.raise_for_status()
        status = response.json().get('status', '')
        if status == "OK" or status == "AlreadySignedIn":
            copilot_enabled = True
        return response.json()
    except requests.RequestException as e:
        raise HTTPException(status_code=500, detail=str(e))

@copilot_app.post("/sign_out",response_model = CopilotPayloadSignOut)
async def sign_in_confirm(req: BlankRequest):
    global copilot_enabled
    try:
        response = requests.post(f"{NODE_SERVER_URL}/sendRequest", json={"method": "signOut", "params": req.dict()})
        response.raise_for_status()
        copilot_enabled = False
        return response.json()
    except requests.RequestException as e:
        raise HTTPException(status_code=500, detail=str(e))
    
async def text_document_did_open(params: DidOpenTextDocumentParams):
    global copilot_enabled
    if copilot_enabled:
        try:
            response = requests.post(f"{NODE_SERVER_URL}/sendNotification", json={"method": "textDocument/didOpen", "params": params.dict()})
            response.raise_for_status()
            return {"message": "Notification sent successfully"}
        except requests.RequestException as e:
            raise HTTPException(status_code=500, detail=str(e))

async def text_document_did_change(params):
    global copilot_enabled
    global copilot_doc_open
    global version
    if copilot_enabled:
        try:
            params = DidChangeTextDocumentParams(**params)
            if not copilot_doc_open:
                open_document = TextDocumentItem(uri=params.textDocument.uri, languageId="python", version=version, text=params.contentChanges[0].text)
                open_request = DidOpenTextDocumentParams(textDocument=open_document)
                response = await text_document_did_open(open_request)
                copilot_doc_open = True
                return response
            else:
                version+=1
                params.textDocument.version = version
                response = requests.post(f"{NODE_SERVER_URL}/sendNotification", json={"method": "textDocument/didChange", "params": params.dict()})
                response.raise_for_status()
                return {"message": "ChangeEvent sent successfully"}
        except requests.RequestException as e:
            raise HTTPException(status_code=500, detail=str(e))

@copilot_app.post("/get_completions", response_model=CopilotGetCompletionsResult)
async def get_completions(params: GetCompletionsParams):
    global copilot_enabled
    global copilot_doc_open
    global version
    if copilot_enabled and copilot_doc_open:
        try:
            params.doc.version = version
            response = requests.post(f"{NODE_SERVER_URL}/sendRequest", json={"method": "getCompletions", "params": params.dict()})
            response.raise_for_status()
            return response.json()
        except requests.RequestException as e:
            raise HTTPException(status_code=500, detail=str(e))
        
@copilot_app.post("/accept_completion")
async def check_status(req: AcceptRequest):
    try:
        response = requests.post(f"{NODE_SERVER_URL}/sendNotification", json={"method": "notifyAccepted", "params": req.dict()})
        return {"message": 'Completion Accepted'}
    except requests.RequestException as e:
        raise HTTPException(status_code=500, detail=str(e))

@copilot_app.post("/reject_completion")
async def check_status(req: RejectRequest):
    try:
        response = requests.post(f"{NODE_SERVER_URL}/sendNotification", json={"method": "notifyRejected", "params": req.dict()})
        return {"message": 'Completion Rejected'}
    except requests.RequestException as e:
        raise HTTPException(status_code=500, detail=str(e))
    


import pandas as pd
from pydantic import Field, BaseModel
import numpy as np
from zt_backend.models.components.zt_component import ZTComponent
from typing import List, Dict,Any


class Header(BaseModel):
    title: str
    align: str = Field("start", description="Column alignment.")
    key: str = 'name'

class DataFrame(ZTComponent):
    component: str = "v-data-table"
    headers: List[Header] 
    items: List[Dict[str, Any]]

    @classmethod
    def from_dataframe(cls, df: pd.DataFrame, id: str):
        df = df.replace({np.nan:None}).replace({np.inf:None}).replace({-np.inf:None})
        headers = [{"title": col, "key": col} for col in df.columns]
        items = df.to_dict(orient='records')
        return cls(id=id, headers=headers, items=items)